@@ -29,8 +29,8 @@ use std::{
     str::{self, FromStr},
 };
 
-use bitcoin::secp256k1;
-use bitcoin::{self, Script};
+use bitcoin::blockdata::witness::Witness;
+use bitcoin::{self, secp256k1, Script};
 
 use self::checksum::verify_checksum;
 use expression;
@@ -41,21 +41,28 @@ use {
     TranslatePk2,
 };
 
+// Directly export from lib.rs, exporting the trait here causes conflicts in this file
+pub(crate) mod pretaproot;
+
 mod bare;
 mod segwitv0;
 mod sh;
 mod sortedmulti;
+mod tr;
+
 // Descriptor Exports
 pub use self::bare::{Bare, Pkh};
 pub use self::segwitv0::{Wpkh, Wsh, WshInner};
 pub use self::sh::{Sh, ShInner};
 pub use self::sortedmulti::SortedMultiVec;
+pub use self::tr::{TapTree, Tr};
 
 mod checksum;
 mod key;
+
 pub use self::key::{
     ConversionError, DescriptorKeyParseError, DescriptorPublicKey, DescriptorSecretKey,
-    DescriptorSinglePriv, DescriptorSinglePub, DescriptorXKey, InnerXKey, Wildcard,
+    DescriptorSinglePriv, DescriptorSinglePub, DescriptorXKey, InnerXKey, SinglePubKey, Wildcard,
 };
 
 /// Alias type for a map of public key to secret key
@@ -85,6 +92,8 @@ pub trait DescriptorTrait<Pk: MiniscriptKey> {
 
     /// Computes the Bitcoin address of the descriptor, if one exists
     /// Some descriptors like pk() don't have any address.
+    /// Errors:
+    /// - On raw/bare descriptors that don't have any address
     fn address(&self, network: bitcoin::Network) -> Result<bitcoin::Address, Error>
     where
         Pk: ToPublicKey;
@@ -110,11 +119,12 @@ pub trait DescriptorTrait<Pk: MiniscriptKey> {
     /// script before any hashing is done. For `Bare`, `Pkh` and `Wpkh` this
     /// is the scriptPubkey; for `ShWpkh` and `Sh` this is the redeemScript;
     /// for the others it is the witness script.
-    fn explicit_script(&self) -> Script
+    /// For `Tr` descriptors, this will error as there is no underlying script
+    fn explicit_script(&self) -> Result<Script, Error>
     where
         Pk: ToPublicKey;
 
-    /// Returns satisfying witness and scriptSig to spend an
+    /// Returns satisfying non-malleable witness and scriptSig with minimum weight to spend an
     /// output controlled by the given descriptor if it possible to
     /// construct one using the satisfier S.
     fn get_satisfaction<S>(&self, satisfier: S) -> Result<(Vec<Vec<u8>>, Script), Error>
@@ -122,7 +132,15 @@ pub trait DescriptorTrait<Pk: MiniscriptKey> {
         Pk: ToPublicKey,
         S: Satisfier<Pk>;
 
-    /// Attempts to produce a satisfying witness and scriptSig to spend an
+    /// Returns satisfying, possibly malleable witness and scriptSig to spend an
+    /// output controlled by the given descriptor if it possible to
+    /// construct one using the satisfier S.
+    fn get_satisfaction_mall<S>(&self, satisfier: S) -> Result<(Vec<Vec<u8>>, Script), Error>
+    where
+        Pk: ToPublicKey,
+        S: Satisfier<Pk>;
+
+    /// Attempts to produce a non-malleable satisfying witness and scriptSig to spend an
     /// output controlled by the given descriptor; add the data to a given
     /// `TxIn` output.
     fn satisfy<S>(&self, txin: &mut bitcoin::TxIn, satisfier: S) -> Result<(), Error>
@@ -132,13 +150,13 @@ pub trait DescriptorTrait<Pk: MiniscriptKey> {
     {
         // easy default implementation
         let (witness, script_sig) = self.get_satisfaction(satisfier)?;
-        txin.witness = witness;
+        txin.witness = Witness::from_vec(witness);
         txin.script_sig = script_sig;
         Ok(())
     }
 
     /// Computes an upper bound on the weight of a satisfying witness to the
-    /// transaction. Assumes all signatures are 73 bytes, including push opcode
+    /// transaction. Assumes all ec-signatures are 73 bytes, including push opcode
     /// and sighash suffix. Includes the weight of the VarInts encoding the
     /// scriptSig and witness stack length.
     /// Returns Error when the descriptor is impossible to safisfy (ex: sh(OP_FALSE))
@@ -148,7 +166,9 @@ pub trait DescriptorTrait<Pk: MiniscriptKey> {
     ///
     /// The `scriptCode` is the Script of the previous transaction output being serialized in the
     /// sighash when evaluating a `CHECKSIG` & co. OP code.
-    fn script_code(&self) -> Script
+    /// Errors:
+    /// - When the descriptor is Tr
+    fn script_code(&self) -> Result<Script, Error>
     where
         Pk: ToPublicKey;
 }
@@ -166,6 +186,50 @@ pub enum Descriptor<Pk: MiniscriptKey> {
     Sh(Sh<Pk>),
     /// Pay-to-Witness-ScriptHash with Segwitv0 context
     Wsh(Wsh<Pk>),
+    /// Pay-to-Taproot
+    Tr(Tr<Pk>),
+}
+
+impl<Pk: MiniscriptKey> From<Bare<Pk>> for Descriptor<Pk> {
+    #[inline]
+    fn from(inner: Bare<Pk>) -> Self {
+        Descriptor::Bare(inner)
+    }
+}
+
+impl<Pk: MiniscriptKey> From<Pkh<Pk>> for Descriptor<Pk> {
+    #[inline]
+    fn from(inner: Pkh<Pk>) -> Self {
+        Descriptor::Pkh(inner)
+    }
+}
+
+impl<Pk: MiniscriptKey> From<Wpkh<Pk>> for Descriptor<Pk> {
+    #[inline]
+    fn from(inner: Wpkh<Pk>) -> Self {
+        Descriptor::Wpkh(inner)
+    }
+}
+
+impl<Pk: MiniscriptKey> From<Sh<Pk>> for Descriptor<Pk> {
+    #[inline]
+    fn from(inner: Sh<Pk>) -> Self {
+        Descriptor::Sh(inner)
+    }
+}
+
+impl<Pk: MiniscriptKey> From<Wsh<Pk>> for Descriptor<Pk> {
+    #[inline]
+    fn from(inner: Wsh<Pk>) -> Self {
+        Descriptor::Wsh(inner)
+    }
+}
+
+impl<Pk: MiniscriptKey> From<Tr<Pk>> for Descriptor<Pk> {
+    #[inline]
+    fn from(inner: Tr<Pk>) -> Self {
+        Descriptor::Tr(inner)
+    }
 }
 
 /// Descriptor Type of the descriptor
@@ -191,6 +255,8 @@ pub enum DescriptorType {
     WshSortedMulti,
     /// Sh Wsh Sorted Multi
     ShWshSortedMulti,
+    /// Tr Descriptor
+    Tr,
 }
 
 impl<Pk: MiniscriptKey> Descriptor<Pk> {
@@ -255,6 +321,18 @@ impl<Pk: MiniscriptKey> Descriptor<Pk> {
         Ok(Descriptor::Bare(Bare::new(ms)?))
     }
 
+    // Wrap with sh
+
+    /// Create a new sh wrapper for the given wpkh descriptor
+    pub fn new_sh_with_wpkh(wpkh: Wpkh<Pk>) -> Self {
+        Descriptor::Sh(Sh::new_with_wpkh(wpkh))
+    }
+
+    /// Create a new sh wrapper for the given wsh descriptor
+    pub fn new_sh_with_wsh(wsh: Wsh<Pk>) -> Self {
+        Descriptor::Sh(Sh::new_with_wsh(wsh))
+    }
+
     // sorted multi
 
     /// Create a new sh sortedmulti descriptor with threshold `k`
@@ -277,6 +355,12 @@ impl<Pk: MiniscriptKey> Descriptor<Pk> {
         Ok(Descriptor::Wsh(Wsh::new_sortedmulti(k, pks)?))
     }
 
+    /// Create new tr descriptor
+    /// Errors when miniscript exceeds resource limits under Tap context
+    pub fn new_tr(key: Pk, script: Option<tr::TapTree<Pk>>) -> Result<Self, Error> {
+        Ok(Descriptor::Tr(Tr::new(key, script)?))
+    }
+
     /// Get the [DescriptorType] of [Descriptor]
     pub fn desc_type(&self) -> DescriptorType {
         match *self {
@@ -296,6 +380,44 @@ impl<Pk: MiniscriptKey> Descriptor<Pk> {
                 WshInner::SortedMulti(ref _smv) => DescriptorType::WshSortedMulti,
                 WshInner::Ms(ref _ms) => DescriptorType::Wsh,
             },
+            Descriptor::Tr(ref _tr) => DescriptorType::Tr,
+        }
+    }
+
+    /// .
+    /// Convert a Descriptor into [`pretaproot::PreTaprootDescriptor`]
+    /// # Examples
+    ///
+    /// ```
+    /// use std::str::FromStr;
+    /// use miniscript::descriptor::Descriptor;
+    /// use miniscript::{PreTaprootDescriptor, PreTaprootDescriptorTrait};
+    /// use miniscript::bitcoin;
+    ///
+    /// // A descriptor with a string generic
+    /// let desc = Descriptor::<bitcoin::PublicKey>::from_str("wpkh(02e18f242c8b0b589bfffeac30e1baa80a60933a649c7fb0f1103e78fbf58aa0ed)")
+    ///     .expect("Valid segwitv0 descriptor");
+    /// let pre_tap_desc = desc.into_pre_taproot_desc().expect("Wsh is pre taproot");
+    ///
+    /// // Now the script code and explicit script no longer fail on longer fail
+    /// // on PreTaprootDescriptor using PreTaprootDescriptorTrait
+    /// let script_code = pre_tap_desc.script_code();
+    /// assert_eq!(script_code.to_string(),
+    ///     "Script(OP_DUP OP_HASH160 OP_PUSHBYTES_20 62107d047e8818b594303fe0657388cc4fc8771f OP_EQUALVERIFY OP_CHECKSIG)"
+    /// );
+    /// ```
+    ///
+    /// # Errors
+    ///
+    /// This function will return an error if descriptor is not a pre taproot descriptor.
+    pub fn into_pre_taproot_desc(self) -> Result<pretaproot::PreTaprootDescriptor<Pk>, Self> {
+        match self {
+            Descriptor::Bare(bare) => Ok(pretaproot::PreTaprootDescriptor::Bare(bare)),
+            Descriptor::Pkh(pkh) => Ok(pretaproot::PreTaprootDescriptor::Pkh(pkh)),
+            Descriptor::Wpkh(wpkh) => Ok(pretaproot::PreTaprootDescriptor::Wpkh(wpkh)),
+            Descriptor::Sh(sh) => Ok(pretaproot::PreTaprootDescriptor::Sh(sh)),
+            Descriptor::Wsh(wsh) => Ok(pretaproot::PreTaprootDescriptor::Wsh(wsh)),
+            Descriptor::Tr(tr) => Err(Descriptor::Tr(tr)),
         }
     }
 }
@@ -332,6 +454,9 @@ impl<P: MiniscriptKey, Q: MiniscriptKey> TranslatePk<P, Q> for Descriptor<P> {
             Descriptor::Wsh(ref wsh) => {
                 Descriptor::Wsh(wsh.translate_pk(&mut translatefpk, &mut translatefpkh)?)
             }
+            Descriptor::Tr(ref tr) => {
+                Descriptor::Tr(tr.translate_pk(&mut translatefpk, &mut translatefpkh)?)
+            }
         };
         Ok(desc)
     }
@@ -353,6 +478,7 @@ impl<Pk: MiniscriptKey> DescriptorTrait<Pk> for Descriptor<Pk> {
             Descriptor::Wpkh(ref wpkh) => wpkh.sanity_check(),
             Descriptor::Wsh(ref wsh) => wsh.sanity_check(),
             Descriptor::Sh(ref sh) => sh.sanity_check(),
+            Descriptor::Tr(ref tr) => tr.sanity_check(),
         }
     }
     /// Computes the Bitcoin address of the descriptor, if one exists
@@ -366,6 +492,7 @@ impl<Pk: MiniscriptKey> DescriptorTrait<Pk> for Descriptor<Pk> {
             Descriptor::Wpkh(ref wpkh) => wpkh.address(network),
             Descriptor::Wsh(ref wsh) => wsh.address(network),
             Descriptor::Sh(ref sh) => sh.address(network),
+            Descriptor::Tr(ref tr) => tr.address(network),
         }
     }
 
@@ -380,6 +507,7 @@ impl<Pk: MiniscriptKey> DescriptorTrait<Pk> for Descriptor<Pk> {
             Descriptor::Wpkh(ref wpkh) => wpkh.script_pubkey(),
             Descriptor::Wsh(ref wsh) => wsh.script_pubkey(),
             Descriptor::Sh(ref sh) => sh.script_pubkey(),
+            Descriptor::Tr(ref tr) => tr.script_pubkey(),
         }
     }
 
@@ -401,6 +529,7 @@ impl<Pk: MiniscriptKey> DescriptorTrait<Pk> for Descriptor<Pk> {
             Descriptor::Wpkh(ref wpkh) => wpkh.unsigned_script_sig(),
             Descriptor::Wsh(ref wsh) => wsh.unsigned_script_sig(),
             Descriptor::Sh(ref sh) => sh.unsigned_script_sig(),
+            Descriptor::Tr(ref tr) => tr.unsigned_script_sig(),
         }
     }
 
@@ -408,7 +537,9 @@ impl<Pk: MiniscriptKey> DescriptorTrait<Pk> for Descriptor<Pk> {
     /// script before any hashing is done. For `Bare`, `Pkh` and `Wpkh` this
     /// is the scriptPubkey; for `ShWpkh` and `Sh` this is the redeemScript;
     /// for the others it is the witness script.
-    fn explicit_script(&self) -> Script
+    /// Errors:
+    /// - When the descriptor is Tr
+    fn explicit_script(&self) -> Result<Script, Error>
     where
         Pk: ToPublicKey,
     {
@@ -418,10 +549,11 @@ impl<Pk: MiniscriptKey> DescriptorTrait<Pk> for Descriptor<Pk> {
             Descriptor::Wpkh(ref wpkh) => wpkh.explicit_script(),
             Descriptor::Wsh(ref wsh) => wsh.explicit_script(),
             Descriptor::Sh(ref sh) => sh.explicit_script(),
+            Descriptor::Tr(ref tr) => tr.explicit_script(),
         }
     }
 
-    /// Returns satisfying witness and scriptSig to spend an
+    /// Returns satisfying non-malleable witness and scriptSig to spend an
     /// output controlled by the given descriptor if it possible to
     /// construct one using the satisfier S.
     fn get_satisfaction<S>(&self, satisfier: S) -> Result<(Vec<Vec<u8>>, Script), Error>
@@ -435,6 +567,25 @@ impl<Pk: MiniscriptKey> DescriptorTrait<Pk> for Descriptor<Pk> {
             Descriptor::Wpkh(ref wpkh) => wpkh.get_satisfaction(satisfier),
             Descriptor::Wsh(ref wsh) => wsh.get_satisfaction(satisfier),
             Descriptor::Sh(ref sh) => sh.get_satisfaction(satisfier),
+            Descriptor::Tr(ref tr) => tr.get_satisfaction(satisfier),
+        }
+    }
+
+    /// Returns a possilbly mallable satisfying non-malleable witness and scriptSig to spend an
+    /// output controlled by the given descriptor if it possible to
+    /// construct one using the satisfier S.
+    fn get_satisfaction_mall<S>(&self, satisfier: S) -> Result<(Vec<Vec<u8>>, Script), Error>
+    where
+        Pk: ToPublicKey,
+        S: Satisfier<Pk>,
+    {
+        match *self {
+            Descriptor::Bare(ref bare) => bare.get_satisfaction_mall(satisfier),
+            Descriptor::Pkh(ref pkh) => pkh.get_satisfaction_mall(satisfier),
+            Descriptor::Wpkh(ref wpkh) => wpkh.get_satisfaction_mall(satisfier),
+            Descriptor::Wsh(ref wsh) => wsh.get_satisfaction_mall(satisfier),
+            Descriptor::Sh(ref sh) => sh.get_satisfaction_mall(satisfier),
+            Descriptor::Tr(ref tr) => tr.get_satisfaction_mall(satisfier),
         }
     }
 
@@ -449,6 +600,7 @@ impl<Pk: MiniscriptKey> DescriptorTrait<Pk> for Descriptor<Pk> {
             Descriptor::Wpkh(ref wpkh) => wpkh.max_satisfaction_weight(),
             Descriptor::Wsh(ref wsh) => wsh.max_satisfaction_weight(),
             Descriptor::Sh(ref sh) => sh.max_satisfaction_weight(),
+            Descriptor::Tr(ref tr) => tr.max_satisfaction_weight(),
         }
     }
 
@@ -456,7 +608,8 @@ impl<Pk: MiniscriptKey> DescriptorTrait<Pk> for Descriptor<Pk> {
     ///
     /// The `scriptCode` is the Script of the previous transaction output being serialized in the
     /// sighash when evaluating a `CHECKSIG` & co. OP code.
-    fn script_code(&self) -> Script
+    /// Returns Error for Tr descriptors
+    fn script_code(&self) -> Result<Script, Error>
     where
         Pk: ToPublicKey,
     {
@@ -466,6 +619,7 @@ impl<Pk: MiniscriptKey> DescriptorTrait<Pk> for Descriptor<Pk> {
             Descriptor::Wpkh(ref wpkh) => wpkh.script_code(),
             Descriptor::Wsh(ref wsh) => wsh.script_code(),
             Descriptor::Sh(ref sh) => sh.script_code(),
+            Descriptor::Tr(ref tr) => tr.script_code(),
         }
     }
 }
@@ -482,6 +636,7 @@ impl<Pk: MiniscriptKey> ForEachKey<Pk> for Descriptor<Pk> {
             Descriptor::Wpkh(ref wpkh) => wpkh.for_each_key(pred),
             Descriptor::Wsh(ref wsh) => wsh.for_each_key(pred),
             Descriptor::Sh(ref sh) => sh.for_each_key(pred),
+            Descriptor::Tr(ref tr) => tr.for_each_key(pred),
         }
     }
 }
@@ -495,10 +650,48 @@ impl Descriptor<DescriptorPublicKey> {
     /// Derives all wildcard keys in the descriptor using the supplied index
     ///
     /// Panics if given an index ≥ 2^31
+    ///
+    /// In most cases, you would want to use [`Self::derived_descriptor`] directly to obtain
+    /// a [`Descriptor<bitcoin::PublicKey>`]
     pub fn derive(&self, index: u32) -> Descriptor<DescriptorPublicKey> {
         self.translate_pk2_infallible(|pk| pk.clone().derive(index))
     }
 
+    /// Derive a [`Descriptor`] with a concrete [`bitcoin::PublicKey`] at a given index
+    /// Removes all extended pubkeys and wildcards from the descriptor and only leaves
+    /// concrete [`bitcoin::PublicKey`]. All [`crate::XOnlyKey`]s are converted to [`bitcoin::PublicKey`]
+    /// by adding a default(0x02) y-coordinate. For [`crate::descriptor::Tr`] descriptor,
+    /// spend info is also cached.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use miniscript::descriptor::{Descriptor, DescriptorPublicKey};
+    /// use miniscript::bitcoin::secp256k1;
+    /// use std::str::FromStr;
+    ///
+    /// // test from bip 86
+    /// let secp = secp256k1::Secp256k1::verification_only();
+    /// let descriptor = Descriptor::<DescriptorPublicKey>::from_str("tr(xpub6BgBgsespWvERF3LHQu6CnqdvfEvtMcQjYrcRzx53QJjSxarj2afYWcLteoGVky7D3UKDP9QyrLprQ3VCECoY49yfdDEHGCtMMj92pReUsQ/0/*)")
+    ///     .expect("Valid ranged descriptor");
+    /// let result = descriptor.derived_descriptor(&secp, 0).expect("Non-hardened derivation");
+    /// assert_eq!(result.to_string(), "tr(03cc8a4bc64d897bddc5fbc2f670f7a8ba0b386779106cf1223c6fc5d7cd6fc115)#6qm9h8ym");
+    /// ```
+    ///
+    /// # Errors
+    ///
+    /// This function will return an error if hardened derivation is attempted.
+    pub fn derived_descriptor<C: secp256k1::Verification>(
+        &self,
+        secp: &secp256k1::Secp256k1<C>,
+        index: u32,
+    ) -> Result<Descriptor<bitcoin::PublicKey>, ConversionError> {
+        let derived = self
+            .derive(index)
+            .translate_pk2(|xpk| xpk.derive_public_key(secp))?;
+        Ok(derived)
+    }
+
     /// Parse a descriptor that may contain secret keys
     ///
     /// Internally turns every secret key found into the corresponding public key and then returns a
@@ -572,6 +765,7 @@ where
             ("wpkh", 1) => Descriptor::Wpkh(Wpkh::from_tree(top)?),
             ("sh", 1) => Descriptor::Sh(Sh::from_tree(top)?),
             ("wsh", 1) => Descriptor::Wsh(Wsh::from_tree(top)?),
+            ("tr", _) => Descriptor::Tr(Tr::from_tree(top)?),
             _ => Descriptor::Bare(Bare::from_tree(top)?),
         })
     }
@@ -587,9 +781,16 @@ where
     type Err = Error;
 
     fn from_str(s: &str) -> Result<Descriptor<Pk>, Error> {
-        let desc_str = verify_checksum(s)?;
-        let top = expression::Tree::from_str(desc_str)?;
-        expression::FromTree::from_tree(&top)
+        // tr tree parsing has special code
+        // Tr::from_str will check the checksum
+        // match "tr(" to handle more extensibly
+        if s.starts_with("tr(") {
+            Ok(Descriptor::Tr(Tr::from_str(s)?))
+        } else {
+            let desc_str = verify_checksum(s)?;
+            let top = expression::Tree::from_str(desc_str)?;
+            expression::FromTree::from_tree(&top)
+        }
     }
 }
 
@@ -601,6 +802,7 @@ impl<Pk: MiniscriptKey> fmt::Debug for Descriptor<Pk> {
             Descriptor::Wpkh(ref wpkh) => write!(f, "{:?}", wpkh),
             Descriptor::Sh(ref sub) => write!(f, "{:?}", sub),
             Descriptor::Wsh(ref sub) => write!(f, "{:?}", sub),
+            Descriptor::Tr(ref tr) => write!(f, "{:?}", tr),
         }
     }
 }
@@ -613,6 +815,7 @@ impl<Pk: MiniscriptKey> fmt::Display for Descriptor<Pk> {
             Descriptor::Wpkh(ref wpkh) => write!(f, "{}", wpkh),
             Descriptor::Sh(ref sub) => write!(f, "{}", sub),
             Descriptor::Wsh(ref sub) => write!(f, "{}", sub),
+            Descriptor::Tr(ref tr) => write!(f, "{}", tr),
         }
     }
 }
@@ -622,20 +825,20 @@ serde_string_impl_pk!(Descriptor, "a script descriptor");
 #[cfg(test)]
 mod tests {
     use super::checksum::desc_checksum;
-    use super::DescriptorTrait;
+    use super::tr::Tr;
+    use super::*;
     use bitcoin::blockdata::opcodes::all::{OP_CLTV, OP_CSV};
     use bitcoin::blockdata::script::Instruction;
     use bitcoin::blockdata::{opcodes, script};
-    use bitcoin::hashes::hex::FromHex;
+    use bitcoin::hashes::hex::{FromHex, ToHex};
     use bitcoin::hashes::{hash160, sha256};
     use bitcoin::util::bip32;
-    use bitcoin::{self, secp256k1, PublicKey};
+    use bitcoin::{self, secp256k1, EcdsaSighashType, PublicKey};
     use descriptor::key::Wildcard;
     use descriptor::{
         DescriptorPublicKey, DescriptorSecretKey, DescriptorSinglePub, DescriptorXKey,
     };
     use hex_script;
-    use miniscript::satisfy::BitcoinSig;
     use std::cmp;
     use std::collections::HashMap;
     use std::str::FromStr;
@@ -908,25 +1111,25 @@ mod tests {
         let secp = secp256k1::Secp256k1::new();
         let sk =
             secp256k1::SecretKey::from_slice(&b"sally was a secret key, she said"[..]).unwrap();
-        let pk = bitcoin::PublicKey {
-            key: secp256k1::PublicKey::from_secret_key(&secp, &sk),
-            compressed: true,
-        };
+        let pk = bitcoin::PublicKey::new(secp256k1::PublicKey::from_secret_key(&secp, &sk));
         let msg = secp256k1::Message::from_slice(&b"michael was a message, amusingly"[..])
             .expect("32 bytes");
-        let sig = secp.sign(&msg, &sk);
+        let sig = secp.sign_ecdsa(&msg, &sk);
         let mut sigser = sig.serialize_der().to_vec();
         sigser.push(0x01); // sighash_all
 
         struct SimpleSat {
-            sig: secp256k1::Signature,
+            sig: secp256k1::ecdsa::Signature,
             pk: bitcoin::PublicKey,
-        };
+        }
 
         impl Satisfier<bitcoin::PublicKey> for SimpleSat {
-            fn lookup_sig(&self, pk: &bitcoin::PublicKey) -> Option<BitcoinSig> {
+            fn lookup_ecdsa_sig(&self, pk: &bitcoin::PublicKey) -> Option<bitcoin::EcdsaSig> {
                 if *pk == self.pk {
-                    Some((self.sig, bitcoin::SigHashType::All))
+                    Some(bitcoin::EcdsaSig {
+                        sig: self.sig,
+                        hash_ty: bitcoin::EcdsaSighashType::All,
+                    })
                 } else {
                     None
                 }
@@ -940,7 +1143,7 @@ mod tests {
             previous_output: bitcoin::OutPoint::default(),
             script_sig: bitcoin::Script::new(),
             sequence: 100,
-            witness: vec![],
+            witness: Witness::default(),
         };
         let bare = Descriptor::new_bare(ms.clone()).unwrap();
 
@@ -951,7 +1154,7 @@ mod tests {
                 previous_output: bitcoin::OutPoint::default(),
                 script_sig: script::Builder::new().push_slice(&sigser[..]).into_script(),
                 sequence: 100,
-                witness: vec![],
+                witness: Witness::default(),
             }
         );
         assert_eq!(bare.unsigned_script_sig(), bitcoin::Script::new());
@@ -967,7 +1170,7 @@ mod tests {
                     .push_key(&pk)
                     .into_script(),
                 sequence: 100,
-                witness: vec![],
+                witness: Witness::default(),
             }
         );
         assert_eq!(pkh.unsigned_script_sig(), bitcoin::Script::new());
@@ -980,7 +1183,7 @@ mod tests {
                 previous_output: bitcoin::OutPoint::default(),
                 script_sig: bitcoin::Script::new(),
                 sequence: 100,
-                witness: vec![sigser.clone(), pk.to_bytes(),],
+                witness: Witness::from_vec(vec![sigser.clone(), pk.to_bytes(),]),
             }
         );
         assert_eq!(wpkh.unsigned_script_sig(), bitcoin::Script::new());
@@ -1001,7 +1204,7 @@ mod tests {
                     .push_slice(&redeem_script[..])
                     .into_script(),
                 sequence: 100,
-                witness: vec![sigser.clone(), pk.to_bytes(),],
+                witness: Witness::from_vec(vec![sigser.clone(), pk.to_bytes(),]),
             }
         );
         assert_eq!(
@@ -1023,7 +1226,7 @@ mod tests {
                     .push_slice(&ms.encode()[..])
                     .into_script(),
                 sequence: 100,
-                witness: vec![],
+                witness: Witness::default(),
             }
         );
         assert_eq!(sh.unsigned_script_sig(), bitcoin::Script::new());
@@ -1038,7 +1241,7 @@ mod tests {
                 previous_output: bitcoin::OutPoint::default(),
                 script_sig: bitcoin::Script::new(),
                 sequence: 100,
-                witness: vec![sigser.clone(), ms.encode().into_bytes(),],
+                witness: Witness::from_vec(vec![sigser.clone(), ms.encode().into_bytes(),]),
             }
         );
         assert_eq!(wsh.unsigned_script_sig(), bitcoin::Script::new());
@@ -1053,7 +1256,7 @@ mod tests {
                     .push_slice(&ms.encode().to_v0_p2wsh()[..])
                     .into_script(),
                 sequence: 100,
-                witness: vec![sigser.clone(), ms.encode().into_bytes(),],
+                witness: Witness::from_vec(vec![sigser.clone(), ms.encode().into_bytes(),]),
             }
         );
         assert_eq!(
@@ -1067,7 +1270,7 @@ mod tests {
     #[test]
     fn after_is_cltv() {
         let descriptor = Descriptor::<bitcoin::PublicKey>::from_str("wsh(after(1000))").unwrap();
-        let script = descriptor.explicit_script();
+        let script = descriptor.explicit_script().unwrap();
 
         let actual_instructions: Vec<_> = script.instructions().collect();
         let check = actual_instructions.last().unwrap();
@@ -1078,7 +1281,7 @@ mod tests {
     #[test]
     fn older_is_csv() {
         let descriptor = Descriptor::<bitcoin::PublicKey>::from_str("wsh(older(1000))").unwrap();
-        let script = descriptor.explicit_script();
+        let script = descriptor.explicit_script().unwrap();
 
         let actual_instructions: Vec<_> = script.instructions().collect();
         let check = actual_instructions.last().unwrap();
@@ -1086,6 +1289,61 @@ mod tests {
         assert_eq!(check, &Ok(Instruction::Op(OP_CSV)))
     }
 
+    #[test]
+    fn tr_roundtrip_key() {
+        let script = Tr::<DummyKey>::from_str("tr()").unwrap().to_string();
+        assert_eq!(script, format!("tr()#x4ml3kxd"))
+    }
+
+    #[test]
+    fn tr_roundtrip_script() {
+        let descriptor = Tr::<DummyKey>::from_str("tr(,{pk(),pk()})")
+            .unwrap()
+            .to_string();
+
+        assert_eq!(descriptor, "tr(,{pk(),pk()})#7dqr6v8r");
+
+        let descriptor = Descriptor::<String>::from_str("tr(A,{pk(B),pk(C)})")
+            .unwrap()
+            .to_string();
+        assert_eq!(descriptor, "tr(A,{pk(B),pk(C)})#y0uc9t6x");
+    }
+
+    #[test]
+    fn tr_roundtrip_tree() {
+        let p1 = "020000000000000000000000000000000000000000000000000000000000000001";
+        let p2 = "020000000000000000000000000000000000000000000000000000000000000002";
+        let p3 = "020000000000000000000000000000000000000000000000000000000000000003";
+        let p4 = "020000000000000000000000000000000000000000000000000000000000000004";
+        let p5 = "f54a5851e9372b87810a8e60cdd2e7cfd80b6e31";
+        let descriptor = Tr::<PublicKey>::from_str(&format!(
+            "tr({},{{pk({}),{{pk({}),or_d(pk({}),pkh({}))}}}})",
+            p1, p2, p3, p4, p5
+        ))
+        .unwrap()
+        .to_string();
+
+        assert_eq!(
+            descriptor,
+            format!(
+                "tr({},{{pk({}),{{pk({}),or_d(pk({}),pkh({}))}}}})#fdhmu4fj",
+                p1, p2, p3, p4, p5
+            )
+        )
+    }
+
+    #[test]
+    fn tr_script_pubkey() {
+        let key = Descriptor::<bitcoin::PublicKey>::from_str(
+            "tr(02e20e746af365e86647826397ba1c0e0d5cb685752976fe2f326ab76bdc4d6ee9)",
+        )
+        .unwrap();
+        assert_eq!(
+            key.script_pubkey().to_hex(),
+            "51209c19294f03757da3dc235a5960631e3c55751632f5889b06b7a053bdc0bcfbcb"
+        )
+    }
+
     #[test]
     fn roundtrip_tests() {
         let descriptor = Descriptor::<bitcoin::PublicKey>::from_str("multi");
@@ -1111,13 +1369,13 @@ mod tests {
             "02937402303919b3a2ee5edd5009f4236f069bf75667b8e6ecf8e5464e20116a0e",
         )
         .unwrap();
-        let sig_a = secp256k1::Signature::from_str("3045022100a7acc3719e9559a59d60d7b2837f9842df30e7edcd754e63227e6168cec72c5d022066c2feba4671c3d99ea75d9976b4da6c86968dbf3bab47b1061e7a1966b1778c").unwrap();
+        let sig_a = secp256k1::ecdsa::Signature::from_str("3045022100a7acc3719e9559a59d60d7b2837f9842df30e7edcd754e63227e6168cec72c5d022066c2feba4671c3d99ea75d9976b4da6c86968dbf3bab47b1061e7a1966b1778c").unwrap();
 
         let b = bitcoin::PublicKey::from_str(
             "02eb64639a17f7334bb5a1a3aad857d6fec65faef439db3de72f85c88bc2906ad3",
         )
         .unwrap();
-        let sig_b = secp256k1::Signature::from_str("3044022075b7b65a7e6cd386132c5883c9db15f9a849a0f32bc680e9986398879a57c276022056d94d12255a4424f51c700ac75122cb354895c9f2f88f0cbb47ba05c9c589ba").unwrap();
+        let sig_b = secp256k1::ecdsa::Signature::from_str("3044022075b7b65a7e6cd386132c5883c9db15f9a849a0f32bc680e9986398879a57c276022056d94d12255a4424f51c700ac75122cb354895c9f2f88f0cbb47ba05c9c589ba").unwrap();
 
         let descriptor = Descriptor::<bitcoin::PublicKey>::from_str(&format!(
             "wsh(and_v(v:pk({A}),pk({B})))",
@@ -1130,13 +1388,25 @@ mod tests {
             previous_output: bitcoin::OutPoint::default(),
             script_sig: bitcoin::Script::new(),
             sequence: 0,
-            witness: vec![],
+            witness: Witness::default(),
         };
         let satisfier = {
             let mut satisfier = HashMap::with_capacity(2);
 
-            satisfier.insert(a, (sig_a.clone(), ::bitcoin::SigHashType::All));
-            satisfier.insert(b, (sig_b.clone(), ::bitcoin::SigHashType::All));
+            satisfier.insert(
+                a,
+                bitcoin::EcdsaSig {
+                    sig: sig_a,
+                    hash_ty: EcdsaSighashType::All,
+                },
+            );
+            satisfier.insert(
+                b,
+                bitcoin::EcdsaSig {
+                    sig: sig_b,
+                    hash_ty: EcdsaSighashType::All,
+                },
+            );
 
             satisfier
         };
@@ -1145,11 +1415,12 @@ mod tests {
         descriptor.satisfy(&mut txin, &satisfier).unwrap();
 
         // assert
-        let witness0 = &txin.witness[0];
-        let witness1 = &txin.witness[1];
+        let wit = txin.witness.to_vec();
+        let witness0 = &wit[0];
+        let witness1 = &wit[1];
 
-        let sig0 = secp256k1::Signature::from_der(&witness0[..witness0.len() - 1]).unwrap();
-        let sig1 = secp256k1::Signature::from_der(&witness1[..witness1.len() - 1]).unwrap();
+        let sig0 = secp256k1::ecdsa::Signature::from_der(&witness0[..witness0.len() - 1]).unwrap();
+        let sig1 = secp256k1::ecdsa::Signature::from_der(&witness1[..witness1.len() - 1]).unwrap();
 
         // why are we asserting this way?
         // The witness stack is evaluated from top to bottom. Given an `and` instruction, the left arm of the and is going to evaluate first,
@@ -1167,7 +1438,7 @@ mod tests {
         )
         .unwrap();
         assert_eq!(
-            *descriptor.script_code().as_bytes(),
+            *descriptor.script_code().unwrap().as_bytes(),
             Vec::<u8>::from_hex("76a9141d0f172a0ecb48aee1be1f2687d2963ae33f71a188ac").unwrap()[..]
         );
 
@@ -1177,7 +1448,7 @@ mod tests {
         )
         .unwrap();
         assert_eq!(
-            *descriptor.script_code().as_bytes(),
+            *descriptor.script_code().unwrap().as_bytes(),
             Vec::<u8>::from_hex("76a91479091972186c449eb1ded22b78e40d009bdf008988ac").unwrap()[..]
         );
 
@@ -1188,7 +1459,7 @@ mod tests {
         .unwrap();
         assert_eq!(
             *descriptor
-                .script_code()
+                .script_code().unwrap()
                 .as_bytes(),
             Vec::<u8>::from_hex("522103789ed0bb717d88f7d321a368d905e7430207ebbd82bd342cf11ae157a7ace5fd2103dbc6764b8884a92e871274b87583e6d5c2a58819473e17e107ef3f6aa5a6162652ae").unwrap()[..]
         );
@@ -1197,7 +1468,7 @@ mod tests {
         let descriptor = Descriptor::<PublicKey>::from_str("sh(wsh(multi(2,03789ed0bb717d88f7d321a368d905e7430207ebbd82bd342cf11ae157a7ace5fd,03dbc6764b8884a92e871274b87583e6d5c2a58819473e17e107ef3f6aa5a61626)))").unwrap();
         assert_eq!(
             *descriptor
-                .script_code()
+                .script_code().unwrap()
                 .as_bytes(),
             Vec::<u8>::from_hex("522103789ed0bb717d88f7d321a368d905e7430207ebbd82bd342cf11ae157a7ace5fd2103dbc6764b8884a92e871274b87583e6d5c2a58819473e17e107ef3f6aa5a6162652ae")
                 .unwrap()[..]
@@ -1261,10 +1532,12 @@ mod tests {
         // Raw (compressed) pubkey
         let key = "03f28773c2d975288bc7d1d205c3748651b075fbc6610e58cddeeddf8f19405aa8";
         let expected = DescriptorPublicKey::SinglePub(DescriptorSinglePub {
-            key: bitcoin::PublicKey::from_str(
-                "03f28773c2d975288bc7d1d205c3748651b075fbc6610e58cddeeddf8f19405aa8",
-            )
-            .unwrap(),
+            key: SinglePubKey::FullKey(
+                bitcoin::PublicKey::from_str(
+                    "03f28773c2d975288bc7d1d205c3748651b075fbc6610e58cddeeddf8f19405aa8",
+                )
+                .unwrap(),
+            ),
             origin: None,
         });
         assert_eq!(expected, key.parse().unwrap());
@@ -1273,10 +1546,10 @@ mod tests {
         // Raw (uncompressed) pubkey
         let key = "04f5eeb2b10c944c6b9fbcfff94c35bdeecd93df977882babc7f3a2cf7f5c81d3b09a68db7f0e04f21de5d4230e75e6dbe7ad16eefe0d4325a62067dc6f369446a";
         let expected = DescriptorPublicKey::SinglePub(DescriptorSinglePub {
-            key: bitcoin::PublicKey::from_str(
+            key: SinglePubKey::FullKey(bitcoin::PublicKey::from_str(
                 "04f5eeb2b10c944c6b9fbcfff94c35bdeecd93df977882babc7f3a2cf7f5c81d3b09a68db7f0e04f21de5d4230e75e6dbe7ad16eefe0d4325a62067dc6f369446a",
             )
-            .unwrap(),
+            .unwrap()),
             origin: None,
         });
         assert_eq!(expected, key.parse().unwrap());
@@ -1286,10 +1559,12 @@ mod tests {
         let desc =
             "[78412e3a/0'/42/0']0231c7d3fc85c148717848033ce276ae2b464a4e2c367ed33886cc428b8af48ff8";
         let expected = DescriptorPublicKey::SinglePub(DescriptorSinglePub {
-            key: bitcoin::PublicKey::from_str(
-                "0231c7d3fc85c148717848033ce276ae2b464a4e2c367ed33886cc428b8af48ff8",
-            )
-            .unwrap(),
+            key: SinglePubKey::FullKey(
+                bitcoin::PublicKey::from_str(
+                    "0231c7d3fc85c148717848033ce276ae2b464a4e2c367ed33886cc428b8af48ff8",
+                )
+                .unwrap(),
+            ),
             origin: Some((
                 bip32::Fingerprint::from(&[0x78, 0x41, 0x2e, 0x3a][..]),
                 (&[
@@ -1446,4 +1721,19 @@ pk(03f28773c2d975288bc7d1d205c3748651b075fbc6610e58cddeeddf8f19405aa8))";
         let descriptor: Descriptor<DescriptorPublicKey> = descriptor_str.parse().unwrap();
         assert_eq!(descriptor.to_string(), "sh(wsh(pk(xpub6ERApfZwUNrhLCkDtcHTcxd75RbzS1ed54G1LkBUHQVHQKqhMkhgbmJbZRkrgZw4koxb5JaHWkY4ALHY2grBGRjaDMzQLcgJvLJuZZvRcEL)))#6c6hwr22");
     }
+
+    #[test]
+    fn test_xonly_keys() {
+        let comp_key = "0308c0fcf8895f4361b4fc77afe2ad53b0bd27dcebfd863421b2b246dc283d4103";
+        let x_only_key = "08c0fcf8895f4361b4fc77afe2ad53b0bd27dcebfd863421b2b246dc283d4103";
+
+        // Both x-only keys and comp keys allowed in tr
+        Descriptor::<DescriptorPublicKey>::from_str(&format!("tr({})", comp_key)).unwrap();
+        Descriptor::<DescriptorPublicKey>::from_str(&format!("tr({})", x_only_key)).unwrap();
+
+        // Only compressed keys allowed in wsh
+        Descriptor::<DescriptorPublicKey>::from_str(&format!("wsh(pk({}))", comp_key)).unwrap();
+        Descriptor::<DescriptorPublicKey>::from_str(&format!("wsh(pk({}))", x_only_key))
+            .unwrap_err();
+    }
 }
