@@ -87,7 +87,6 @@
 //! ```
 //!
 //!
-#![allow(bare_trait_objects)]
 #![cfg_attr(all(test, feature = "unstable"), feature(test))]
 // Coding conventions
 #![deny(unsafe_code)]
@@ -117,6 +116,7 @@ pub mod psbt;
 
 mod util;
 
+use std::io::Write;
 use std::str::FromStr;
 use std::{error, fmt, hash, str};
 
@@ -125,11 +125,13 @@ use bitcoin::hashes::{hash160, sha256, Hash};
 
 pub use descriptor::{Descriptor, DescriptorPublicKey, DescriptorTrait};
 pub use interpreter::Interpreter;
-pub use miniscript::context::{BareCtx, Legacy, ScriptContext, Segwitv0};
+pub use miniscript::context::{BareCtx, Legacy, ScriptContext, Segwitv0, Tap};
 pub use miniscript::decode::Terminal;
-pub use miniscript::satisfy::{BitcoinSig, Preimage32, Satisfier};
+pub use miniscript::satisfy::{Preimage32, Satisfier};
 pub use miniscript::Miniscript;
 
+pub use descriptor::pretaproot::{traits::PreTaprootDescriptorTrait, PreTaprootDescriptor};
+
 ///Public key trait which can be converted to Hash type
 pub trait MiniscriptKey: Clone + Eq + Ord + fmt::Debug + fmt::Display + hash::Hash {
     /// Check if the publicKey is uncompressed. The default
@@ -137,20 +139,37 @@ pub trait MiniscriptKey: Clone + Eq + Ord + fmt::Debug + fmt::Display + hash::Ha
     fn is_uncompressed(&self) -> bool {
         false
     }
+
+    /// Check if the publicKey is x-only. The default
+    /// implementation returns false
+    //
+    // This is required to know what in DescriptorPublicKey to know whether the inner
+    // key in allowed in descriptor context
+    fn is_x_only_key(&self) -> bool {
+        false
+    }
+
     /// The associated Hash type with the publicKey
     type Hash: Clone + Eq + Ord + fmt::Display + fmt::Debug + hash::Hash;
 
     /// Converts an object to PublicHash
     fn to_pubkeyhash(&self) -> Self::Hash;
+}
 
-    /// Computes the size of a public key when serialized in a script,
-    /// including the length bytes
-    fn serialized_len(&self) -> usize {
-        if self.is_uncompressed() {
-            66
-        } else {
-            34
-        }
+impl MiniscriptKey for bitcoin::secp256k1::PublicKey {
+    /// `is_uncompressed` always returns `false`
+    fn is_uncompressed(&self) -> bool {
+        false
+    }
+
+    type Hash = hash160::Hash;
+
+    fn to_pubkeyhash(&self) -> Self::Hash {
+        let mut engine = hash160::Hash::engine();
+        engine
+            .write_all(&self.serialize())
+            .expect("engines don't error");
+        hash160::Hash::from_engine(engine)
     }
 }
 
@@ -170,6 +189,18 @@ impl MiniscriptKey for bitcoin::PublicKey {
     }
 }
 
+impl MiniscriptKey for bitcoin::secp256k1::XOnlyPublicKey {
+    type Hash = hash160::Hash;
+
+    fn to_pubkeyhash(&self) -> Self::Hash {
+        hash160::Hash::hash(&self.serialize())
+    }
+
+    fn is_x_only_key(&self) -> bool {
+        true
+    }
+}
+
 impl MiniscriptKey for String {
     type Hash = String;
 
@@ -183,6 +214,12 @@ pub trait ToPublicKey: MiniscriptKey {
     /// Converts an object to a public key
     fn to_public_key(&self) -> bitcoin::PublicKey;
 
+    /// Convert an object to x-only pubkey
+    fn to_x_only_pubkey(&self) -> bitcoin::secp256k1::XOnlyPublicKey {
+        let pk = self.to_public_key();
+        bitcoin::secp256k1::XOnlyPublicKey::from(pk.inner)
+    }
+
     /// Converts a hashed version of the public key to a `hash160` hash.
     ///
     /// This method must be consistent with `to_public_key`, in the sense
@@ -202,6 +239,35 @@ impl ToPublicKey for bitcoin::PublicKey {
     }
 }
 
+impl ToPublicKey for bitcoin::secp256k1::PublicKey {
+    fn to_public_key(&self) -> bitcoin::PublicKey {
+        bitcoin::PublicKey::new(*self)
+    }
+
+    fn hash_to_hash160(hash: &hash160::Hash) -> hash160::Hash {
+        *hash
+    }
+}
+
+impl ToPublicKey for bitcoin::secp256k1::XOnlyPublicKey {
+    fn to_public_key(&self) -> bitcoin::PublicKey {
+        // This code should never be used.
+        // But is implemented for completeness
+        let mut data: Vec<u8> = vec![0x02];
+        data.extend(self.serialize().iter());
+        bitcoin::PublicKey::from_slice(&data)
+            .expect("Failed to construct 33 Publickey from 0x02 appended x-only key")
+    }
+
+    fn to_x_only_pubkey(&self) -> bitcoin::secp256k1::XOnlyPublicKey {
+        *self
+    }
+
+    fn hash_to_hash160(hash: &hash160::Hash) -> hash160::Hash {
+        *hash
+    }
+}
+
 /// Dummy key which de/serializes to the empty string; useful sometimes for testing
 #[derive(Copy, Clone, PartialOrd, Ord, PartialEq, Eq, Debug)]
 pub struct DummyKey;
@@ -448,13 +514,17 @@ pub enum Error {
     InvalidOpcode(opcodes::All),
     /// Some opcode occurred followed by `OP_VERIFY` when it had
     /// a `VERIFY` version that should have been used instead
-    NonMinimalVerify(miniscript::lex::Token),
+    NonMinimalVerify(String),
     /// Push was illegal in some context
     InvalidPush(Vec<u8>),
     /// rust-bitcoin script error
     Script(script::Error),
+    /// rust-bitcoin address error
+    AddrError(bitcoin::util::address::Error),
     /// A `CHECKMULTISIG` opcode was preceded by a number > 20
     CmsTooManyKeys(u32),
+    /// A tapscript multi_a cannot support more than MAX_BLOCK_WEIGHT/32 keys
+    MultiATooManyKeys(u32),
     /// Encountered unprintable character in descriptor
     Unprintable(u8),
     /// expected character while parsing descriptor; didn't find one
@@ -469,10 +539,6 @@ pub enum Error {
     MultiAt(String),
     /// Name of a fragment contained `@` but we were not parsing an OR
     AtOutsideOr(String),
-    /// Fragment was an `and_v(_, true)` which should be written as `t:`
-    NonCanonicalTrue,
-    /// Fragment was an `or_i(_, false)` or `or_i(false,_)` which should be written as `u:` or `l:`
-    NonCanonicalFalse,
     /// Encountered a `l:0` which is syntactically equal to `u:0` except stupid
     LikelyFalse,
     /// Encountered a wrapping character that we don't recognize
@@ -521,6 +587,14 @@ pub enum Error {
     ImpossibleSatisfaction,
     /// Bare descriptors don't have any addresses
     BareDescriptorAddr,
+    /// PubKey invalid under current context
+    PubKeyCtxError(miniscript::decode::KeyParseError, &'static str),
+    /// Attempted to call function that requires PreComputed taproot info
+    TaprootSpendInfoUnavialable,
+    /// No script code for Tr descriptors
+    TrNoScriptCode,
+    /// No explicit script for Tr descriptors
+    TrNoExplicitScript,
 }
 
 #[doc(hidden)]
@@ -562,12 +636,19 @@ impl From<bitcoin::secp256k1::Error> for Error {
     }
 }
 
+#[doc(hidden)]
+impl From<bitcoin::util::address::Error> for Error {
+    fn from(e: bitcoin::util::address::Error) -> Error {
+        Error::AddrError(e)
+    }
+}
+
 fn errstr(s: &str) -> Error {
     Error::Unexpected(s.to_owned())
 }
 
 impl error::Error for Error {
-    fn cause(&self) -> Option<&error::Error> {
+    fn cause(&self) -> Option<&dyn error::Error> {
         match *self {
             Error::BadPubkey(ref e) => Some(e),
             _ => None,
@@ -584,9 +665,10 @@ impl fmt::Display for Error {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         match *self {
             Error::InvalidOpcode(op) => write!(f, "invalid opcode {}", op),
-            Error::NonMinimalVerify(tok) => write!(f, "{} VERIFY", tok),
+            Error::NonMinimalVerify(ref tok) => write!(f, "{} VERIFY", tok),
             Error::InvalidPush(ref push) => write!(f, "invalid push {:?}", push), // TODO hexify this
             Error::Script(ref e) => fmt::Display::fmt(e, f),
+            Error::AddrError(ref e) => fmt::Display::fmt(e, f),
             Error::CmsTooManyKeys(n) => write!(f, "checkmultisig with {} keys", n),
             Error::Unprintable(x) => write!(f, "unprintable character 0x{:02x}", x),
             Error::ExpectedChar(c) => write!(f, "expected {}", c),
@@ -595,10 +677,6 @@ impl fmt::Display for Error {
             Error::MultiColon(ref s) => write!(f, "«{}» has multiple instances of «:»", s),
             Error::MultiAt(ref s) => write!(f, "«{}» has multiple instances of «@»", s),
             Error::AtOutsideOr(ref s) => write!(f, "«{}» contains «@» in non-or() context", s),
-            Error::NonCanonicalTrue => f.write_str("Use «t:X» rather than «and_v(X,true())»"),
-            Error::NonCanonicalFalse => {
-                f.write_str("Use «u:X» «l:X» rather than «or_i(X,false)» «or_i(false,X)»")
-            }
             Error::LikelyFalse => write!(f, "0 is not very likely (use «u:0»)"),
             Error::UnknownWrapper(ch) => write!(f, "unknown wrapper «{}:»", ch),
             Error::NonTopLevel(ref s) => write!(f, "non-T miniscript: {}", s),
@@ -642,6 +720,21 @@ impl fmt::Display for Error {
             Error::AnalysisError(ref e) => e.fmt(f),
             Error::ImpossibleSatisfaction => write!(f, "Impossible to satisfy Miniscript"),
             Error::BareDescriptorAddr => write!(f, "Bare descriptors don't have address"),
+            Error::PubKeyCtxError(ref pk, ref ctx) => {
+                write!(f, "Pubkey error: {} under {} scriptcontext", pk, ctx)
+            }
+            Error::MultiATooManyKeys(k) => {
+                write!(f, "MultiA too many keys {}", k)
+            }
+            Error::TaprootSpendInfoUnavialable => {
+                write!(f, "Taproot Spend Info not computed. Hint: Did you call `compute_spend_info` before calling methods from DescriptorTrait")
+            }
+            Error::TrNoScriptCode => {
+                write!(f, "No script code for Tr descriptors")
+            }
+            Error::TrNoExplicitScript => {
+                write!(f, "No script code for Tr descriptors")
+            }
         }
     }
 }
