@@ -18,14 +18,16 @@
 //! scriptpubkeys.
 //!
 
-use std::collections::HashMap;
+use std::collections::{BTreeMap, HashMap};
 use std::sync::Arc;
 use std::{cmp, i64, mem};
 
+use bitcoin;
 use bitcoin::hashes::{hash160, ripemd160, sha256, sha256d};
-use bitcoin::{self, secp256k1};
+use bitcoin::util::taproot::{ControlBlock, LeafVersion, TapLeafHash};
 use {MiniscriptKey, ToPublicKey};
 
+use bitcoin::secp256k1::XOnlyPublicKey;
 use miniscript::limits::{
     HEIGHT_TIME_THRESHOLD, SEQUENCE_LOCKTIME_DISABLE_FLAG, SEQUENCE_LOCKTIME_TYPE_FLAG,
 };
@@ -34,28 +36,32 @@ use Miniscript;
 use ScriptContext;
 use Terminal;
 
-/// Type alias for a signature/hashtype pair
-pub type BitcoinSig = (secp256k1::Signature, bitcoin::SigHashType);
 /// Type alias for 32 byte Preimage.
 pub type Preimage32 = [u8; 32];
-
-/// Helper function to create BitcoinSig from Rawsig
-/// Useful for downstream when implementing Satisfier.
-/// Returns underlying secp if the Signature is not of correct format
-pub fn bitcoinsig_from_rawsig(rawsig: &[u8]) -> Result<BitcoinSig, ::interpreter::Error> {
-    let (flag, sig) = rawsig.split_last().unwrap();
-    let flag = bitcoin::SigHashType::from_u32_standard(*flag as u32)
-        .map_err(|_| ::interpreter::Error::NonStandardSigHash([sig, &[*flag]].concat().to_vec()))?;
-    let sig = secp256k1::Signature::from_der(sig)?;
-    Ok((sig, flag))
-}
 /// Trait describing a lookup table for signatures, hash preimages, etc.
 /// Every method has a default implementation that simply returns `None`
 /// on every query. Users are expected to override the methods that they
 /// have data for.
 pub trait Satisfier<Pk: MiniscriptKey + ToPublicKey> {
-    /// Given a public key, look up a signature with that key
-    fn lookup_sig(&self, _: &Pk) -> Option<BitcoinSig> {
+    /// Given a public key, look up an ECDSA signature with that key
+    fn lookup_ecdsa_sig(&self, _: &Pk) -> Option<bitcoin::EcdsaSig> {
+        None
+    }
+
+    /// Lookup the tap key spend sig
+    fn lookup_tap_key_spend_sig(&self) -> Option<bitcoin::SchnorrSig> {
+        None
+    }
+
+    /// Given a public key and a associated leaf hash, look up an schnorr signature with that key
+    fn lookup_tap_leaf_script_sig(&self, _: &Pk, _: &TapLeafHash) -> Option<bitcoin::SchnorrSig> {
+        None
+    }
+
+    /// Obtain a reference to the control block for a ver and script
+    fn lookup_tap_control_block_map(
+        &self,
+    ) -> Option<&BTreeMap<ControlBlock, (bitcoin::Script, LeafVersion)>> {
         None
     }
 
@@ -64,11 +70,25 @@ pub trait Satisfier<Pk: MiniscriptKey + ToPublicKey> {
         None
     }
 
-    /// Given a keyhash, look up the signature and the associated key
+    /// Given a keyhash, look up the EC signature and the associated key
     /// Even if signatures for public key Hashes are not available, the users
     /// can use this map to provide pkh -> pk mapping which can be useful
     /// for dissatisfying pkh.
-    fn lookup_pkh_sig(&self, _: &Pk::Hash) -> Option<(bitcoin::PublicKey, BitcoinSig)> {
+    fn lookup_pkh_ecdsa_sig(
+        &self,
+        _: &Pk::Hash,
+    ) -> Option<(bitcoin::PublicKey, bitcoin::EcdsaSig)> {
+        None
+    }
+
+    /// Given a keyhash, look up the schnorr signature and the associated key
+    /// Even if signatures for public key Hashes are not available, the users
+    /// can use this map to provide pkh -> pk mapping which can be useful
+    /// for dissatisfying pkh.
+    fn lookup_pkh_tap_leaf_script_sig(
+        &self,
+        _: &(Pk::Hash, TapLeafHash),
+    ) -> Option<(XOnlyPublicKey, bitcoin::SchnorrSig)> {
         None
     }
 
@@ -146,17 +166,29 @@ impl<Pk: MiniscriptKey + ToPublicKey> Satisfier<Pk> for After {
     }
 }
 
-impl<Pk: MiniscriptKey + ToPublicKey> Satisfier<Pk> for HashMap<Pk, BitcoinSig> {
-    fn lookup_sig(&self, key: &Pk) -> Option<BitcoinSig> {
+impl<Pk: MiniscriptKey + ToPublicKey> Satisfier<Pk> for HashMap<Pk, bitcoin::EcdsaSig> {
+    fn lookup_ecdsa_sig(&self, key: &Pk) -> Option<bitcoin::EcdsaSig> {
         self.get(key).map(|x| *x)
     }
 }
 
-impl<Pk: MiniscriptKey + ToPublicKey> Satisfier<Pk> for HashMap<Pk::Hash, (Pk, BitcoinSig)>
+impl<Pk: MiniscriptKey + ToPublicKey> Satisfier<Pk>
+    for HashMap<(Pk, TapLeafHash), bitcoin::SchnorrSig>
+{
+    fn lookup_tap_leaf_script_sig(&self, key: &Pk, h: &TapLeafHash) -> Option<bitcoin::SchnorrSig> {
+        // Unfortunately, there is no way to get a &(a, b) from &a and &b without allocating
+        // If we change the signature the of lookup_tap_leaf_script_sig to accept a tuple. We would
+        // face the same problem while satisfying PkK.
+        // We use this signature to optimize for the psbt common use case.
+        self.get(&(key.clone(), *h)).map(|x| *x)
+    }
+}
+
+impl<Pk: MiniscriptKey + ToPublicKey> Satisfier<Pk> for HashMap<Pk::Hash, (Pk, bitcoin::EcdsaSig)>
 where
     Pk: MiniscriptKey + ToPublicKey,
 {
-    fn lookup_sig(&self, key: &Pk) -> Option<BitcoinSig> {
+    fn lookup_ecdsa_sig(&self, key: &Pk) -> Option<bitcoin::EcdsaSig> {
         self.get(&key.to_pubkeyhash()).map(|x| x.1)
     }
 
@@ -164,23 +196,68 @@ where
         self.get(pk_hash).map(|x| x.0.clone())
     }
 
-    fn lookup_pkh_sig(&self, pk_hash: &Pk::Hash) -> Option<(bitcoin::PublicKey, BitcoinSig)> {
+    fn lookup_pkh_ecdsa_sig(
+        &self,
+        pk_hash: &Pk::Hash,
+    ) -> Option<(bitcoin::PublicKey, bitcoin::EcdsaSig)> {
         self.get(pk_hash)
             .map(|&(ref pk, sig)| (pk.to_public_key(), sig))
     }
 }
 
+impl<Pk: MiniscriptKey + ToPublicKey> Satisfier<Pk>
+    for HashMap<(Pk::Hash, TapLeafHash), (Pk, bitcoin::SchnorrSig)>
+where
+    Pk: MiniscriptKey + ToPublicKey,
+{
+    fn lookup_tap_leaf_script_sig(&self, key: &Pk, h: &TapLeafHash) -> Option<bitcoin::SchnorrSig> {
+        self.get(&(key.to_pubkeyhash(), *h)).map(|x| x.1)
+    }
+
+    fn lookup_pkh_tap_leaf_script_sig(
+        &self,
+        pk_hash: &(Pk::Hash, TapLeafHash),
+    ) -> Option<(XOnlyPublicKey, bitcoin::SchnorrSig)> {
+        self.get(pk_hash)
+            .map(|&(ref pk, sig)| (pk.to_x_only_pubkey(), sig))
+    }
+}
+
 impl<'a, Pk: MiniscriptKey + ToPublicKey, S: Satisfier<Pk>> Satisfier<Pk> for &'a S {
-    fn lookup_sig(&self, p: &Pk) -> Option<BitcoinSig> {
-        (**self).lookup_sig(p)
+    fn lookup_ecdsa_sig(&self, p: &Pk) -> Option<bitcoin::EcdsaSig> {
+        (**self).lookup_ecdsa_sig(p)
+    }
+
+    fn lookup_tap_leaf_script_sig(&self, p: &Pk, h: &TapLeafHash) -> Option<bitcoin::SchnorrSig> {
+        (**self).lookup_tap_leaf_script_sig(p, h)
     }
 
     fn lookup_pkh_pk(&self, pkh: &Pk::Hash) -> Option<Pk> {
         (**self).lookup_pkh_pk(pkh)
     }
 
-    fn lookup_pkh_sig(&self, pkh: &Pk::Hash) -> Option<(bitcoin::PublicKey, BitcoinSig)> {
-        (**self).lookup_pkh_sig(pkh)
+    fn lookup_pkh_ecdsa_sig(
+        &self,
+        pkh: &Pk::Hash,
+    ) -> Option<(bitcoin::PublicKey, bitcoin::EcdsaSig)> {
+        (**self).lookup_pkh_ecdsa_sig(pkh)
+    }
+
+    fn lookup_tap_key_spend_sig(&self) -> Option<bitcoin::SchnorrSig> {
+        (**self).lookup_tap_key_spend_sig()
+    }
+
+    fn lookup_pkh_tap_leaf_script_sig(
+        &self,
+        pkh: &(Pk::Hash, TapLeafHash),
+    ) -> Option<(XOnlyPublicKey, bitcoin::SchnorrSig)> {
+        (**self).lookup_pkh_tap_leaf_script_sig(pkh)
+    }
+
+    fn lookup_tap_control_block_map(
+        &self,
+    ) -> Option<&BTreeMap<ControlBlock, (bitcoin::Script, LeafVersion)>> {
+        (**self).lookup_tap_control_block_map()
     }
 
     fn lookup_sha256(&self, h: sha256::Hash) -> Option<Preimage32> {
@@ -209,16 +286,40 @@ impl<'a, Pk: MiniscriptKey + ToPublicKey, S: Satisfier<Pk>> Satisfier<Pk> for &'
 }
 
 impl<'a, Pk: MiniscriptKey + ToPublicKey, S: Satisfier<Pk>> Satisfier<Pk> for &'a mut S {
-    fn lookup_sig(&self, p: &Pk) -> Option<BitcoinSig> {
-        (**self).lookup_sig(p)
+    fn lookup_ecdsa_sig(&self, p: &Pk) -> Option<bitcoin::EcdsaSig> {
+        (**self).lookup_ecdsa_sig(p)
+    }
+
+    fn lookup_tap_leaf_script_sig(&self, p: &Pk, h: &TapLeafHash) -> Option<bitcoin::SchnorrSig> {
+        (**self).lookup_tap_leaf_script_sig(p, h)
+    }
+
+    fn lookup_tap_key_spend_sig(&self) -> Option<bitcoin::SchnorrSig> {
+        (**self).lookup_tap_key_spend_sig()
     }
 
     fn lookup_pkh_pk(&self, pkh: &Pk::Hash) -> Option<Pk> {
         (**self).lookup_pkh_pk(pkh)
     }
 
-    fn lookup_pkh_sig(&self, pkh: &Pk::Hash) -> Option<(bitcoin::PublicKey, BitcoinSig)> {
-        (**self).lookup_pkh_sig(pkh)
+    fn lookup_pkh_ecdsa_sig(
+        &self,
+        pkh: &Pk::Hash,
+    ) -> Option<(bitcoin::PublicKey, bitcoin::EcdsaSig)> {
+        (**self).lookup_pkh_ecdsa_sig(pkh)
+    }
+
+    fn lookup_pkh_tap_leaf_script_sig(
+        &self,
+        pkh: &(Pk::Hash, TapLeafHash),
+    ) -> Option<(XOnlyPublicKey, bitcoin::SchnorrSig)> {
+        (**self).lookup_pkh_tap_leaf_script_sig(pkh)
+    }
+
+    fn lookup_tap_control_block_map(
+        &self,
+    ) -> Option<&BTreeMap<ControlBlock, (bitcoin::Script, LeafVersion)>> {
+        (**self).lookup_tap_control_block_map()
     }
 
     fn lookup_sha256(&self, h: sha256::Hash) -> Option<Preimage32> {
@@ -254,23 +355,56 @@ macro_rules! impl_tuple_satisfier {
             Pk: MiniscriptKey + ToPublicKey,
             $($ty: Satisfier< Pk>,)*
         {
-            fn lookup_sig(&self, key: &Pk) -> Option<BitcoinSig> {
+            fn lookup_ecdsa_sig(&self, key: &Pk) -> Option<bitcoin::EcdsaSig> {
                 let &($(ref $ty,)*) = self;
                 $(
-                    if let Some(result) = $ty.lookup_sig(key) {
+                    if let Some(result) = $ty.lookup_ecdsa_sig(key) {
                         return Some(result);
                     }
                 )*
                 None
             }
 
-            fn lookup_pkh_sig(
-                &self,
-                key_hash: &Pk::Hash,
-            ) -> Option<(bitcoin::PublicKey, BitcoinSig)> {
+            fn lookup_tap_key_spend_sig(&self) -> Option<bitcoin::SchnorrSig> {
                 let &($(ref $ty,)*) = self;
                 $(
-                    if let Some(result) = $ty.lookup_pkh_sig(key_hash) {
+                    if let Some(result) = $ty.lookup_tap_key_spend_sig() {
+                        return Some(result);
+                    }
+                )*
+                None
+            }
+
+            fn lookup_tap_leaf_script_sig(&self, key: &Pk, h: &TapLeafHash) -> Option<bitcoin::SchnorrSig> {
+                let &($(ref $ty,)*) = self;
+                $(
+                    if let Some(result) = $ty.lookup_tap_leaf_script_sig(key, h) {
+                        return Some(result);
+                    }
+                )*
+                None
+            }
+
+            fn lookup_pkh_ecdsa_sig(
+                &self,
+                key_hash: &Pk::Hash,
+            ) -> Option<(bitcoin::PublicKey, bitcoin::EcdsaSig)> {
+                let &($(ref $ty,)*) = self;
+                $(
+                    if let Some(result) = $ty.lookup_pkh_ecdsa_sig(key_hash) {
+                        return Some(result);
+                    }
+                )*
+                None
+            }
+
+            fn lookup_pkh_tap_leaf_script_sig(
+                &self,
+                key_hash: &(Pk::Hash, TapLeafHash),
+            ) -> Option<(XOnlyPublicKey, bitcoin::SchnorrSig)> {
+                let &($(ref $ty,)*) = self;
+                $(
+                    if let Some(result) = $ty.lookup_pkh_tap_leaf_script_sig(key_hash) {
                         return Some(result);
                     }
                 )*
@@ -290,6 +424,18 @@ macro_rules! impl_tuple_satisfier {
                 None
             }
 
+            fn lookup_tap_control_block_map(
+                &self,
+            ) -> Option<&BTreeMap<ControlBlock, (bitcoin::Script, LeafVersion)>> {
+                let &($(ref $ty,)*) = self;
+                $(
+                    if let Some(result) = $ty.lookup_tap_control_block_map() {
+                        return Some(result);
+                    }
+                )*
+                None
+            }
+
             fn lookup_sha256(&self, h: sha256::Hash) -> Option<Preimage32> {
                 let &($(ref $ty,)*) = self;
                 $(
@@ -401,15 +547,23 @@ impl Ord for Witness {
 
 impl Witness {
     /// Turn a signature into (part of) a satisfaction
-    fn signature<Pk: ToPublicKey, S: Satisfier<Pk>>(sat: S, pk: &Pk) -> Self {
-        match sat.lookup_sig(pk) {
-            Some((sig, hashtype)) => {
-                let mut ret = sig.serialize_der().to_vec();
-                ret.push(hashtype.as_u32() as u8);
-                Witness::Stack(vec![ret])
-            }
-            // Signatures cannot be forged
-            None => Witness::Impossible,
+    fn signature<Pk: ToPublicKey, S: Satisfier<Pk>, Ctx: ScriptContext>(
+        sat: S,
+        pk: &Pk,
+        leaf_hash: &TapLeafHash,
+    ) -> Self {
+        match Ctx::sig_type() {
+            super::context::SigType::Ecdsa => match sat.lookup_ecdsa_sig(pk) {
+                Some(sig) => Witness::Stack(vec![sig.to_vec()]),
+                // Signatures cannot be forged
+                None => Witness::Impossible,
+            },
+            super::context::SigType::Schnorr => match sat.lookup_tap_leaf_script_sig(pk, leaf_hash)
+            {
+                Some(sig) => Witness::Stack(vec![sig.to_vec()]),
+                // Signatures cannot be forged
+                None => Witness::Impossible,
+            },
         }
     }
 
@@ -425,12 +579,8 @@ impl Witness {
 
     /// Turn a key/signature pair related to a pkh into (part of) a satisfaction
     fn pkh_signature<Pk: ToPublicKey, S: Satisfier<Pk>>(sat: S, pkh: &Pk::Hash) -> Self {
-        match sat.lookup_pkh_sig(pkh) {
-            Some((pk, (sig, hashtype))) => {
-                let mut ret = sig.serialize_der().to_vec();
-                ret.push(hashtype.as_u32() as u8);
-                Witness::Stack(vec![ret.to_vec(), pk.to_public_key().to_bytes()])
-            }
+        match sat.lookup_pkh_ecdsa_sig(pkh) {
+            Some((pk, sig)) => Witness::Stack(vec![sig.to_vec(), pk.to_public_key().to_bytes()]),
             None => Witness::Impossible,
         }
     }
@@ -523,6 +673,7 @@ impl Satisfaction {
         subs: &[Arc<Miniscript<Pk, Ctx>>],
         stfr: &Sat,
         root_has_sig: bool,
+        leaf_hash: &TapLeafHash,
         min_fn: &mut F,
     ) -> Self
     where
@@ -533,13 +684,29 @@ impl Satisfaction {
     {
         let mut sats = subs
             .iter()
-            .map(|s| Self::satisfy_helper(&s.node, stfr, root_has_sig, min_fn, &mut Self::thresh))
+            .map(|s| {
+                Self::satisfy_helper(
+                    &s.node,
+                    stfr,
+                    root_has_sig,
+                    leaf_hash,
+                    min_fn,
+                    &mut Self::thresh,
+                )
+            })
             .collect::<Vec<_>>();
         // Start with the to-return stack set to all dissatisfactions
         let mut ret_stack = subs
             .iter()
             .map(|s| {
-                Self::dissatisfy_helper(&s.node, stfr, root_has_sig, min_fn, &mut Self::thresh)
+                Self::dissatisfy_helper(
+                    &s.node,
+                    stfr,
+                    root_has_sig,
+                    leaf_hash,
+                    min_fn,
+                    &mut Self::thresh,
+                )
             })
             .collect::<Vec<_>>();
 
@@ -624,6 +791,7 @@ impl Satisfaction {
         subs: &[Arc<Miniscript<Pk, Ctx>>],
         stfr: &Sat,
         root_has_sig: bool,
+        leaf_hash: &TapLeafHash,
         min_fn: &mut F,
     ) -> Self
     where
@@ -635,14 +803,28 @@ impl Satisfaction {
         let mut sats = subs
             .iter()
             .map(|s| {
-                Self::satisfy_helper(&s.node, stfr, root_has_sig, min_fn, &mut Self::thresh_mall)
+                Self::satisfy_helper(
+                    &s.node,
+                    stfr,
+                    root_has_sig,
+                    leaf_hash,
+                    min_fn,
+                    &mut Self::thresh_mall,
+                )
             })
             .collect::<Vec<_>>();
         // Start with the to-return stack set to all dissatisfactions
         let mut ret_stack = subs
             .iter()
             .map(|s| {
-                Self::dissatisfy_helper(&s.node, stfr, root_has_sig, min_fn, &mut Self::thresh_mall)
+                Self::dissatisfy_helper(
+                    &s.node,
+                    stfr,
+                    root_has_sig,
+                    leaf_hash,
+                    min_fn,
+                    &mut Self::thresh_mall,
+                )
             })
             .collect::<Vec<_>>();
 
@@ -737,6 +919,7 @@ impl Satisfaction {
         term: &Terminal<Pk, Ctx>,
         stfr: &Sat,
         root_has_sig: bool,
+        leaf_hash: &TapLeafHash,
         min_fn: &mut F,
         thresh_fn: &mut G,
     ) -> Self
@@ -745,11 +928,18 @@ impl Satisfaction {
         Ctx: ScriptContext,
         Sat: Satisfier<Pk>,
         F: FnMut(Satisfaction, Satisfaction) -> Satisfaction,
-        G: FnMut(usize, &[Arc<Miniscript<Pk, Ctx>>], &Sat, bool, &mut F) -> Satisfaction,
+        G: FnMut(
+            usize,
+            &[Arc<Miniscript<Pk, Ctx>>],
+            &Sat,
+            bool,
+            &TapLeafHash,
+            &mut F,
+        ) -> Satisfaction,
     {
         match *term {
             Terminal::PkK(ref pk) => Satisfaction {
-                stack: Witness::signature(stfr, pk),
+                stack: Witness::signature::<_, _, Ctx>(stfr, pk, leaf_hash),
                 has_sig: true,
             },
             Terminal::PkH(ref pkh) => Satisfaction {
@@ -817,29 +1007,47 @@ impl Satisfaction {
             | Terminal::Verify(ref sub)
             | Terminal::NonZero(ref sub)
             | Terminal::ZeroNotEqual(ref sub) => {
-                Self::satisfy_helper(&sub.node, stfr, root_has_sig, min_fn, thresh_fn)
+                Self::satisfy_helper(&sub.node, stfr, root_has_sig, leaf_hash, min_fn, thresh_fn)
             }
             Terminal::DupIf(ref sub) => {
-                let sat = Self::satisfy_helper(&sub.node, stfr, root_has_sig, min_fn, thresh_fn);
+                let sat = Self::satisfy_helper(
+                    &sub.node,
+                    stfr,
+                    root_has_sig,
+                    leaf_hash,
+                    min_fn,
+                    thresh_fn,
+                );
                 Satisfaction {
                     stack: Witness::combine(sat.stack, Witness::push_1()),
                     has_sig: sat.has_sig,
                 }
             }
             Terminal::AndV(ref l, ref r) | Terminal::AndB(ref l, ref r) => {
-                let l_sat = Self::satisfy_helper(&l.node, stfr, root_has_sig, min_fn, thresh_fn);
-                let r_sat = Self::satisfy_helper(&r.node, stfr, root_has_sig, min_fn, thresh_fn);
+                let l_sat =
+                    Self::satisfy_helper(&l.node, stfr, root_has_sig, leaf_hash, min_fn, thresh_fn);
+                let r_sat =
+                    Self::satisfy_helper(&r.node, stfr, root_has_sig, leaf_hash, min_fn, thresh_fn);
                 Satisfaction {
                     stack: Witness::combine(r_sat.stack, l_sat.stack),
                     has_sig: l_sat.has_sig || r_sat.has_sig,
                 }
             }
             Terminal::AndOr(ref a, ref b, ref c) => {
-                let a_sat = Self::satisfy_helper(&a.node, stfr, root_has_sig, min_fn, thresh_fn);
-                let a_nsat =
-                    Self::dissatisfy_helper(&a.node, stfr, root_has_sig, min_fn, thresh_fn);
-                let b_sat = Self::satisfy_helper(&b.node, stfr, root_has_sig, min_fn, thresh_fn);
-                let c_sat = Self::satisfy_helper(&c.node, stfr, root_has_sig, min_fn, thresh_fn);
+                let a_sat =
+                    Self::satisfy_helper(&a.node, stfr, root_has_sig, leaf_hash, min_fn, thresh_fn);
+                let a_nsat = Self::dissatisfy_helper(
+                    &a.node,
+                    stfr,
+                    root_has_sig,
+                    leaf_hash,
+                    min_fn,
+                    thresh_fn,
+                );
+                let b_sat =
+                    Self::satisfy_helper(&b.node, stfr, root_has_sig, leaf_hash, min_fn, thresh_fn);
+                let c_sat =
+                    Self::satisfy_helper(&c.node, stfr, root_has_sig, leaf_hash, min_fn, thresh_fn);
 
                 min_fn(
                     Satisfaction {
@@ -853,12 +1061,26 @@ impl Satisfaction {
                 )
             }
             Terminal::OrB(ref l, ref r) => {
-                let l_sat = Self::satisfy_helper(&l.node, stfr, root_has_sig, min_fn, thresh_fn);
-                let r_sat = Self::satisfy_helper(&r.node, stfr, root_has_sig, min_fn, thresh_fn);
-                let l_nsat =
-                    Self::dissatisfy_helper(&l.node, stfr, root_has_sig, min_fn, thresh_fn);
-                let r_nsat =
-                    Self::dissatisfy_helper(&r.node, stfr, root_has_sig, min_fn, thresh_fn);
+                let l_sat =
+                    Self::satisfy_helper(&l.node, stfr, root_has_sig, leaf_hash, min_fn, thresh_fn);
+                let r_sat =
+                    Self::satisfy_helper(&r.node, stfr, root_has_sig, leaf_hash, min_fn, thresh_fn);
+                let l_nsat = Self::dissatisfy_helper(
+                    &l.node,
+                    stfr,
+                    root_has_sig,
+                    leaf_hash,
+                    min_fn,
+                    thresh_fn,
+                );
+                let r_nsat = Self::dissatisfy_helper(
+                    &r.node,
+                    stfr,
+                    root_has_sig,
+                    leaf_hash,
+                    min_fn,
+                    thresh_fn,
+                );
 
                 assert!(!l_nsat.has_sig);
                 assert!(!r_nsat.has_sig);
@@ -875,10 +1097,18 @@ impl Satisfaction {
                 )
             }
             Terminal::OrD(ref l, ref r) | Terminal::OrC(ref l, ref r) => {
-                let l_sat = Self::satisfy_helper(&l.node, stfr, root_has_sig, min_fn, thresh_fn);
-                let r_sat = Self::satisfy_helper(&r.node, stfr, root_has_sig, min_fn, thresh_fn);
-                let l_nsat =
-                    Self::dissatisfy_helper(&l.node, stfr, root_has_sig, min_fn, thresh_fn);
+                let l_sat =
+                    Self::satisfy_helper(&l.node, stfr, root_has_sig, leaf_hash, min_fn, thresh_fn);
+                let r_sat =
+                    Self::satisfy_helper(&r.node, stfr, root_has_sig, leaf_hash, min_fn, thresh_fn);
+                let l_nsat = Self::dissatisfy_helper(
+                    &l.node,
+                    stfr,
+                    root_has_sig,
+                    leaf_hash,
+                    min_fn,
+                    thresh_fn,
+                );
 
                 assert!(!l_nsat.has_sig);
 
@@ -891,8 +1121,10 @@ impl Satisfaction {
                 )
             }
             Terminal::OrI(ref l, ref r) => {
-                let l_sat = Self::satisfy_helper(&l.node, stfr, root_has_sig, min_fn, thresh_fn);
-                let r_sat = Self::satisfy_helper(&r.node, stfr, root_has_sig, min_fn, thresh_fn);
+                let l_sat =
+                    Self::satisfy_helper(&l.node, stfr, root_has_sig, leaf_hash, min_fn, thresh_fn);
+                let r_sat =
+                    Self::satisfy_helper(&r.node, stfr, root_has_sig, leaf_hash, min_fn, thresh_fn);
                 min_fn(
                     Satisfaction {
                         stack: Witness::combine(l_sat.stack, Witness::push_1()),
@@ -904,13 +1136,15 @@ impl Satisfaction {
                     },
                 )
             }
-            Terminal::Thresh(k, ref subs) => thresh_fn(k, subs, stfr, root_has_sig, min_fn),
+            Terminal::Thresh(k, ref subs) => {
+                thresh_fn(k, subs, stfr, root_has_sig, leaf_hash, min_fn)
+            }
             Terminal::Multi(k, ref keys) => {
                 // Collect all available signatures
                 let mut sig_count = 0;
                 let mut sigs = Vec::with_capacity(k);
                 for pk in keys {
-                    match Witness::signature(stfr, pk) {
+                    match Witness::signature::<_, _, Ctx>(stfr, pk, leaf_hash) {
                         Witness::Stack(sig) => {
                             sigs.push(sig);
                             sig_count += 1;
@@ -947,6 +1181,44 @@ impl Satisfaction {
                     }
                 }
             }
+            Terminal::MultiA(k, ref keys) => {
+                // Collect all available signatures
+                let mut sig_count = 0;
+                let mut sigs = vec![vec![vec![]]; keys.len()];
+                for (i, pk) in keys.iter().rev().enumerate() {
+                    match Witness::signature::<_, _, Ctx>(stfr, pk, leaf_hash) {
+                        Witness::Stack(sig) => {
+                            sigs[i] = sig;
+                            sig_count += 1;
+                            // This a privacy issue, we are only selecting the first available
+                            // sigs. Incase pk at pos 1 is not selected, we know we did not have access to it
+                            // bitcoin core also implements the same logic for MULTISIG, so I am not bothering
+                            // permuting the sigs for now
+                            if sig_count == k {
+                                break;
+                            }
+                        }
+                        Witness::Impossible => {}
+                        Witness::Unavailable => unreachable!(
+                            "Signature satisfaction without witness must be impossible"
+                        ),
+                    }
+                }
+
+                if sig_count < k {
+                    Satisfaction {
+                        stack: Witness::Impossible,
+                        has_sig: false,
+                    }
+                } else {
+                    Satisfaction {
+                        stack: sigs.into_iter().fold(Witness::empty(), |acc, sig| {
+                            Witness::combine(acc, Witness::Stack(sig))
+                        }),
+                        has_sig: true,
+                    }
+                }
+            }
         }
     }
 
@@ -955,6 +1227,7 @@ impl Satisfaction {
         term: &Terminal<Pk, Ctx>,
         stfr: &Sat,
         root_has_sig: bool,
+        leaf_hash: &TapLeafHash,
         min_fn: &mut F,
         thresh_fn: &mut G,
     ) -> Self
@@ -963,7 +1236,14 @@ impl Satisfaction {
         Ctx: ScriptContext,
         Sat: Satisfier<Pk>,
         F: FnMut(Satisfaction, Satisfaction) -> Satisfaction,
-        G: FnMut(usize, &[Arc<Miniscript<Pk, Ctx>>], &Sat, bool, &mut F) -> Satisfaction,
+        G: FnMut(
+            usize,
+            &[Arc<Miniscript<Pk, Ctx>>],
+            &Sat,
+            bool,
+            &TapLeafHash,
+            &mut F,
+        ) -> Satisfaction,
     {
         match *term {
             Terminal::PkK(..) => Satisfaction {
@@ -1001,7 +1281,7 @@ impl Satisfaction {
             | Terminal::Swap(ref sub)
             | Terminal::Check(ref sub)
             | Terminal::ZeroNotEqual(ref sub) => {
-                Self::dissatisfy_helper(&sub.node, stfr, root_has_sig, min_fn, thresh_fn)
+                Self::dissatisfy_helper(&sub.node, stfr, root_has_sig, leaf_hash, min_fn, thresh_fn)
             }
             Terminal::DupIf(_) | Terminal::NonZero(_) => Satisfaction {
                 stack: Witness::push_0(),
@@ -1012,9 +1292,16 @@ impl Satisfaction {
                 has_sig: false,
             },
             Terminal::AndV(ref v, ref other) => {
-                let vsat = Self::satisfy_helper(&v.node, stfr, root_has_sig, min_fn, thresh_fn);
-                let odissat =
-                    Self::dissatisfy_helper(&other.node, stfr, root_has_sig, min_fn, thresh_fn);
+                let vsat =
+                    Self::satisfy_helper(&v.node, stfr, root_has_sig, leaf_hash, min_fn, thresh_fn);
+                let odissat = Self::dissatisfy_helper(
+                    &other.node,
+                    stfr,
+                    root_has_sig,
+                    leaf_hash,
+                    min_fn,
+                    thresh_fn,
+                );
                 Satisfaction {
                     stack: Witness::combine(odissat.stack, vsat.stack),
                     has_sig: vsat.has_sig || odissat.has_sig,
@@ -1024,8 +1311,22 @@ impl Satisfaction {
             | Terminal::OrB(ref l, ref r)
             | Terminal::OrD(ref l, ref r)
             | Terminal::AndOr(ref l, _, ref r) => {
-                let lnsat = Self::dissatisfy_helper(&l.node, stfr, root_has_sig, min_fn, thresh_fn);
-                let rnsat = Self::dissatisfy_helper(&r.node, stfr, root_has_sig, min_fn, thresh_fn);
+                let lnsat = Self::dissatisfy_helper(
+                    &l.node,
+                    stfr,
+                    root_has_sig,
+                    leaf_hash,
+                    min_fn,
+                    thresh_fn,
+                );
+                let rnsat = Self::dissatisfy_helper(
+                    &r.node,
+                    stfr,
+                    root_has_sig,
+                    leaf_hash,
+                    min_fn,
+                    thresh_fn,
+                );
                 Satisfaction {
                     stack: Witness::combine(rnsat.stack, lnsat.stack),
                     has_sig: rnsat.has_sig || lnsat.has_sig,
@@ -1036,24 +1337,45 @@ impl Satisfaction {
                 has_sig: false,
             },
             Terminal::OrI(ref l, ref r) => {
-                let lnsat = Self::dissatisfy_helper(&l.node, stfr, root_has_sig, min_fn, thresh_fn);
+                let lnsat = Self::dissatisfy_helper(
+                    &l.node,
+                    stfr,
+                    root_has_sig,
+                    leaf_hash,
+                    min_fn,
+                    thresh_fn,
+                );
                 let dissat_1 = Satisfaction {
                     stack: Witness::combine(lnsat.stack, Witness::push_1()),
                     has_sig: lnsat.has_sig,
                 };
 
-                let rnsat = Self::dissatisfy_helper(&r.node, stfr, root_has_sig, min_fn, thresh_fn);
+                let rnsat = Self::dissatisfy_helper(
+                    &r.node,
+                    stfr,
+                    root_has_sig,
+                    leaf_hash,
+                    min_fn,
+                    thresh_fn,
+                );
                 let dissat_2 = Satisfaction {
                     stack: Witness::combine(rnsat.stack, Witness::push_0()),
                     has_sig: rnsat.has_sig,
                 };
 
-                min_fn(dissat_1, dissat_2)
+                // Dissatisfactions don't need to non-malleable. Use minimum_mall always
+                Satisfaction::minimum_mall(dissat_1, dissat_2)
             }
             Terminal::Thresh(_, ref subs) => Satisfaction {
                 stack: subs.iter().fold(Witness::empty(), |acc, sub| {
-                    let nsat =
-                        Self::dissatisfy_helper(&sub.node, stfr, root_has_sig, min_fn, thresh_fn);
+                    let nsat = Self::dissatisfy_helper(
+                        &sub.node,
+                        stfr,
+                        root_has_sig,
+                        leaf_hash,
+                        min_fn,
+                        thresh_fn,
+                    );
                     assert!(!nsat.has_sig);
                     Witness::combine(nsat.stack, acc)
                 }),
@@ -1063,6 +1385,10 @@ impl Satisfaction {
                 stack: Witness::Stack(vec![vec![]; k + 1]),
                 has_sig: false,
             },
+            Terminal::MultiA(_, ref pks) => Satisfaction {
+                stack: Witness::Stack(vec![vec![]; pks.len()]),
+                has_sig: false,
+            },
         }
     }
 
@@ -1075,11 +1401,13 @@ impl Satisfaction {
         term: &Terminal<Pk, Ctx>,
         stfr: &Sat,
         root_has_sig: bool,
+        leaf_hash: &TapLeafHash,
     ) -> Self {
         Self::satisfy_helper(
             term,
             stfr,
             root_has_sig,
+            leaf_hash,
             &mut Satisfaction::minimum,
             &mut Satisfaction::thresh,
         )
@@ -1094,11 +1422,13 @@ impl Satisfaction {
         term: &Terminal<Pk, Ctx>,
         stfr: &Sat,
         root_has_sig: bool,
+        leaf_hash: &TapLeafHash,
     ) -> Self {
         Self::satisfy_helper(
             term,
             stfr,
             root_has_sig,
+            leaf_hash,
             &mut Satisfaction::minimum_mall,
             &mut Satisfaction::thresh_mall,
         )
